from dataclasses import dataclass, field
from pathlib import Path
from typing import List


@dataclass
class ReportOutput:
    def __init__(self, path: Path, name: str = None):
        self.path = Path(path)
        self.name = name


@dataclass
class ReportResult:
    name: str = None
    info: str = None  # optional extra info about this report to display to the user
    output_figures: List[ReportOutput] = field(default_factory=lambda: [])
    output_tables: List[ReportOutput] = field(default_factory=lambda: [])
    output_text: List[ReportOutput] = field(default_factory=lambda: [])



from ligo.dsl.symbol_table.SymbolTable import SymbolTable
from ligo.dsl.symbol_table.SymbolType import SymbolType


class DefinitionParserOutput:

    def __init__(self, symbol_table: SymbolTable, specification: dict):
        assert any(len(symbol_table.get_by_type(symbol_type)) > 0 for symbol_type in [t for t in SymbolType]), \
            "DefinitionParserOutput: symbol table has not been populated by objects of any type."

        self.symbol_table = symbol_table
        self.specification = specification



import copy
import os
import random
from dataclasses import fields, field
from itertools import chain
from multiprocessing import Pool
from pathlib import Path
from typing import List, Dict, Tuple

import dill
import math
import numpy as np
from bionumpy.bnpdataclass import BNPDataClass

from ligo.IO.dataset_export.AIRRExporter import AIRRExporter
from ligo.app.LigoApp import SimError
from ligo.data_model.dataset.ReceptorDataset import ReceptorDataset
from ligo.data_model.dataset.RepertoireDataset import RepertoireDataset
from ligo.data_model.dataset.SequenceDataset import SequenceDataset
from ligo.data_model.receptor.Receptor import Receptor
from ligo.data_model.receptor.ReceptorBuilder import ReceptorBuilder
from ligo.data_model.repertoire.Repertoire import Repertoire
from ligo.environment.SequenceType import SequenceType
from ligo.simulation.LigoSimState import LigoSimState
from ligo.simulation.SimConfig import SimConfig
from ligo.simulation.SimConfigItem import SimConfigItem
from ligo.simulation.generative_models.BackgroundSequences import BackgroundSequences
from ligo.simulation.implants.Signal import Signal
from ligo.simulation.simulation_strategy.ImplantingStrategy import ImplantingStrategy
from ligo.simulation.util.bnp_util import merge_dataclass_objects
from ligo.simulation.util.util import get_bnp_data, make_receptor_sequence_objects, make_annotated_dataclass, \
    get_sequence_per_signal_count, \
    update_seqs_without_signal, update_seqs_with_signal, check_iteration_progress, make_sequence_paths, \
    make_signal_metadata, needs_seqs_with_signal, \
    check_sequence_count, make_repertoire_from_sequences, get_no_signal_sequences, get_signal_sequences, \
    annotate_sequences, get_signal_sequence_count, filter_sequences_by_length
from ligo.util.ExporterHelper import ExporterHelper
from ligo.util.Logger import print_log
from ligo.util.PathBuilder import PathBuilder
from ligo.workflows.instructions.Instruction import Instruction
from ligo.workflows.instructions.ligo_simulation.runtime_reports import make_p_gen_histogram_plot


class LigoSimInstruction(Instruction):
    """
    LIgO simulation instruction creates a synthetic dataset from scratch based on the generative model and a set of signals provided by
    the user.

    Arguments:

    - simulation (str): a name of a simulation object containing a list of SimConfigItem as specified under definitions key; defines how to combine signals with simulated data; specified under definitions

    - sequence_batch_size (int): how many sequences to generate at once using the generative model before checking for signals and filtering

    - max_iterations (int): how many iterations are allowed when creating sequences

    - export_p_gens (bool): whether to compute generation probabilities (if supported by the generative model) for sequences and include them as part of output

    - number_of_processes (int): determines how many simulation items can be simulated in parallel

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_simulation_instruction: # user-defined name of the instruction
            type: LIgOSim # which instruction to execute
            simulation: sim1
            sequence_batch_size: 1000
            max_iterations: 1000
            export_p_gens: False
            number_of_processes: 4

    """

    def __init__(self, simulation: SimConfig, signals: List[Signal], name: str,
                 sequence_batch_size: int, max_iterations: int, number_of_processes: int, export_p_gens: bool = None):

        self.state = LigoSimState(simulation=simulation, signals=signals, name=name)
        self._number_of_processes = number_of_processes
        self._sequence_batch_size = sequence_batch_size
        self._max_iterations = max_iterations
        self._export_p_gens = export_p_gens

        self._use_p_gens = self.state.simulation.keep_p_gen_dist and \
                           all(sim_item.generative_model.can_compute_p_gens() for sim_item in
                               self.state.simulation.sim_items)

        self._export_observed_signals = any(
            [it.false_negative_prob_in_receptors > 0 or it.false_positive_prob_in_receptors > 0
             for it in self.state.simulation.sim_items])

        self._noise_fields = [(f"observed_{s.id}", int) for s in
                              self.state.signals] if self._export_observed_signals else []

        if isinstance(self.state.simulation.simulation_strategy, ImplantingStrategy):
            field_seq, field_p_gen = field(default='', metadata=None), field(default=-1., metadata=None)
            implanting_fields = [('original_sequence', str, dill.dumps(field_seq)),
                                 ('original_p_gen', float, dill.dumps(field_p_gen))]
        else:
            implanting_fields = []

        self._annotation_fields = sorted(
            [(signal.id, int) for signal in self.state.signals] + [('signals_aggregated', str)] +
            [(f"{signal.id}_positions", str) for signal in self.state.signals] + self._noise_fields,
            key=lambda x: x[0]) + implanting_fields

        self._custom_fields = self._annotation_fields + [('p_gen', float), ('from_default_model', int)]
        self._background_fields = [(f.name, f.type) for f in fields(BackgroundSequences)]

    @property
    def sequence_type(self) -> SequenceType:
        return self.state.simulation.sequence_type

    @property
    def _annotated_dataclass(self):
        return make_annotated_dataclass(self._annotation_fields, self.state.signals)

    MIN_RANGE_PROBABILITY = 1e-5

    def run(self, result_path: Path):
        self.state.result_path = PathBuilder.build(result_path / self.state.name)

        self._simulate_dataset()
        self._export_dataset()

        return self.state

    def _export_dataset(self):

        exporter_output = ExporterHelper.export_dataset(self.state.resulting_dataset, [AIRRExporter],
                                                        self.state.result_path, omit_columns=['from_default_model'])

        self.state.formats = exporter_output['formats']
        self.state.paths = exporter_output['paths']

    def _simulate_dataset(self):

        examples = self._create_examples_wrapper()
        random.shuffle(examples)

        labels = {**{signal.id: [True, False] for signal in self.state.signals},
                  **{'species': self.state.simulation.species}}

        if self.state.simulation.is_repertoire:
            self.state.resulting_dataset = RepertoireDataset.build_from_objects(labels=labels, repertoires=examples,
                                                                                name='simulated_dataset',
                                                                                metadata_path=self.state.result_path / 'metadata.csv')
        elif self.state.simulation.paired:
            self.state.resulting_dataset = ReceptorDataset.build_from_objects(examples, path=self.state.result_path,
                                                                              name='simulated_dataset',
                                                                              file_size=SequenceDataset.DEFAULT_FILE_SIZE,
                                                                              labels=labels)
        else:
            self.state.resulting_dataset = SequenceDataset.build_from_objects(examples, path=self.state.result_path,
                                                                              name='simulated_dataset',
                                                                              file_size=SequenceDataset.DEFAULT_FILE_SIZE,
                                                                              labels=labels)

    def _create_examples_wrapper(self) -> list:
        if self._number_of_processes > 1:
            chunk_size = math.ceil(len(self.state.simulation.sim_items) / self._number_of_processes)

            with Pool(processes=max(self._number_of_processes, len(self.state.simulation.sim_items))) as pool:
                result = pool.map(self._create_examples, [dill.dumps(item) for item in self.state.simulation.sim_items],
                                  chunksize=chunk_size)
                examples = {k: v for d in result for k, v in d.items()}
        else:
            examples = {}
            for item in self.state.simulation.sim_items:
                examples = {**examples, **self._create_examples(item)}

        if self.state.simulation.paired:
            examples = self._pair_examples(examples, self.state.result_path / 'paired')
        else:
            examples = list(chain.from_iterable(examples.values()))

        return examples

    def _pair_examples(self, examples: Dict[str, list], path: Path) -> list:
        paired_examples = []
        pair_func = self._pair_repertoires if self.state.simulation.is_repertoire else self._pair_sequences
        for paired_item1, paired_item2 in self.state.simulation.paired:
            paired_examples.extend(pair_func(examples[paired_item1], examples[paired_item2], path))
        return paired_examples

    def _pair_repertoires(self, repertoires1: list, repertoires2: list, path: Path) -> List[Repertoire]:
        assert len(repertoires1) == len(
            repertoires2), f"{LigoSimInstruction.__name__}: cannot create paired repertoires, number of repertoires per chain don't match: {len(repertoires1)} and {len(repertoires2)}."
        PathBuilder.build(path)
        paired_repertoires = []
        for i in range(len(repertoires1)):
            repertoire = self._pair_two_repertories(repertoires1[i], repertoires2[i], path)
            paired_repertoires.append(repertoire)
        return paired_repertoires

    def _pair_two_repertories(self, repertoire1: Repertoire, repertoire2: Repertoire, path: Path) -> Repertoire:
        assert repertoire1.get_element_count() == repertoire2.get_element_count(), f"{LigoSimInstruction.__name__}: cannot pair repertoires {repertoire1.identifier} and {repertoire2.identifier}, they have different number of sequences: {repertoire1.get_element_count()} and {repertoire2.get_element_count()}."

        sequences = []
        sequences1, sequences2 = repertoire1.sequences, repertoire2.sequences
        for index, seq1, seq2 in zip(list(range(len(sequences1))), sequences1, sequences2):
            seq1.metadata.cell_id = index
            seq2.metadata.cell_id = index
            seq1.identifier = f"{seq1.identifier}_{seq1.metadata.chain.value}"
            seq2.identifier = f"{seq2.identifier}_{seq2.metadata.chain.value}"
            sequences.extend([seq1, seq2])

        return Repertoire.build_from_sequence_objects(sequences, path,
                                                      metadata={**repertoire1.metadata, **repertoire2.metadata})

    def _pair_sequences(self, sequences1: list, sequences2: list, path: Path = None) -> List[Receptor]:
        assert len(sequences1) == len(
            sequences2), f"{LigoSimInstruction.__name__}: could not create paired dataset, the number of sequences in two simulation items did not match."

        random.shuffle(sequences1)
        random.shuffle(sequences2)

        return ReceptorBuilder.build_objects_from_pairs(sequences1, sequences2)

    def _create_examples(self, item_in) -> Dict[str, list]:

        item = dill.loads(item_in) if isinstance(item_in, bytes) else item_in

        if self.state.simulation.is_repertoire:
            res = self._create_repertoires(item)
        else:
            res = self._create_receptors(item)

        return {item.name: res}

    def _create_receptors(self, sim_item: SimConfigItem) -> list:

        assert len(sim_item.signals) in [0,
                                         1], f"{LigoSimInstruction.__name__}: for sequence datasets, only 0 or 1 signal or a signal pair per " \
                                             f"sequence are supported, but {len(sim_item.signals)} were specified."

        sequence_paths = self._gen_necessary_sequences(self.state.result_path, sim_item)
        sequences = None

        if len(sim_item.signals) == 0:
            sequences = get_bnp_data(sequence_paths['no_signal'], self._annotated_dataclass)
        else:
            for signal in sim_item.signals:
                signal_sequences = get_bnp_data(sequence_paths[signal.id], self._annotated_dataclass)
                sequences = signal_sequences if sequences is None else merge_dataclass_objects(
                    [sequences, signal_sequences])

        sequences = self._compute_p_gens_for_export(sequences, sim_item)

        sequences = make_receptor_sequence_objects(sequences,
                                                   metadata=make_signal_metadata(sim_item, self.state.signals),
                                                   immune_events=sim_item.immune_events,
                                                   custom_params=self._custom_fields,
                                                   chain=sim_item.generative_model.chain)

        return sequences

    def _compute_p_gens_for_export(self, sequences, sim_item: SimConfigItem):
        if self._export_p_gens:
            sequences = self._update_sequences_with_missing_p_gens(sequences, sim_item)
            if isinstance(self.state.simulation.simulation_strategy, ImplantingStrategy):
                sequences = self._update_sequences_with_missing_p_gens(sequences, sim_item,
                                                                       'original_sequence', 'original_p_gen')

        return sequences

    def _create_repertoires(self, item: SimConfigItem) -> list:
        path = PathBuilder.build(self.state.result_path / item.name)
        sequence_paths = self._gen_necessary_sequences(path, sim_item=item)

        repertoires = []
        used_seq_count = {**{'no_signal': 0}, **{signal.id: 0 for signal in item.signals}}
        repertoires_path = PathBuilder.build(path / "repertoires")

        for i in range(item.number_of_examples):
            seqs_no_signal_count = item.receptors_in_repertoire_count - sum(
                get_signal_sequence_count(1, proportion, item.receptors_in_repertoire_count)
                for _, proportion in item.signal_proportions.items())

            sequences, used_seq_count = get_signal_sequences(self._annotated_dataclass, used_seq_count, item,
                                                             sequence_paths)

            sequences, used_seq_count = get_no_signal_sequences(sequences=sequences, used_seq_count=used_seq_count,
                                                                seqs_no_signal_count=seqs_no_signal_count,
                                                                bnp_data_class=self._annotated_dataclass,
                                                                sequence_paths=sequence_paths,
                                                                sim_item=item)

            check_sequence_count(item, sequences)

            sequences = self._compute_p_gens_for_export(sequences, item)

            repertoire = make_repertoire_from_sequences(sequences, repertoires_path, item, self.state.signals,
                                                        self._custom_fields)
            repertoires.append(repertoire)

        return repertoires

    def _gen_necessary_sequences(self, base_path: Path, sim_item: SimConfigItem) -> Dict[str, Path]:

        if sim_item.seed is not None:
            np.random.seed(sim_item.seed)

        path = PathBuilder.build(base_path / sim_item.name)
        seqs_per_signal_count = get_sequence_per_signal_count(sim_item)
        seq_paths = make_sequence_paths(path, sim_item.signals)
        iteration = 1

        while sum(seqs_per_signal_count.values()) > 0 and iteration < self._max_iterations:
            sequences = self._make_background_sequences(path, iteration, sim_item, seqs_per_signal_count,
                                                        need_background_seqs=iteration == 1 and self.state.simulation.keep_p_gen_dist)

            if self.state.simulation.keep_p_gen_dist and sim_item.generative_model.can_compute_p_gens() and iteration == 1:
                self._make_p_gen_histogram(sequences, sim_item.name, path)
                print_log(
                    f"Computed a histogram from the first batch of background sequences for {sim_item.name}, available at: {str(path)}",
                    include_datetime=True)

            sequences = filter_sequences_by_length(sequences, sim_item, self.sequence_type)

            sequences = annotate_sequences(sequences, self.sequence_type == SequenceType.AMINO_ACID, self.state.signals,
                                           self._annotated_dataclass, sim_item.name)

            sequences = self.state.simulation.simulation_strategy.process_sequences(sequences, copy.deepcopy(
                seqs_per_signal_count), self._use_p_gens,  self.sequence_type, sim_item, self.state.signals,
                self.state.simulation.remove_seqs_with_signals,
                implanting_scaling_factor=self.state.simulation.implanting_scaling_factor)

            if sequences is not None and len(sequences) > 0:

                if self.state.simulation.keep_p_gen_dist and sim_item.generative_model.can_compute_p_gens():
                    sequences = self._filter_using_p_gens(sequences, sim_item)

                seqs_per_signal_count['no_signal'] = update_seqs_without_signal(seqs_per_signal_count['no_signal'],
                                                                                sequences, seq_paths['no_signal'])
                seqs_per_signal_count = update_seqs_with_signal(copy.deepcopy(seqs_per_signal_count), sequences,
                                                                self.state.signals, sim_item.signals,
                                                                seq_paths)

            print_log(
                f"Finished iteration {iteration} in {sim_item.name}: remaining sequence count per signal for {sim_item.name}: "
                f"{seqs_per_signal_count}" if sum(
                    seqs_per_signal_count.values()) > 0 else f"{sim_item.name} simulation finished", True)
            check_iteration_progress(iteration, self._max_iterations)
            iteration += 1

        if iteration == self._max_iterations and sum(seqs_per_signal_count.values()) != 0:
            raise SimError(
                f"{LigoSimInstruction.__name__}: maximum iterations were reached, but the simulation could not finish "
                f"with parameters: {vars(self.state.simulation)}.\n")

        return seq_paths

    def _make_background_sequences(self, path, iteration: int, sim_item: SimConfigItem, sequence_per_signal_count: dict,
                                   need_background_seqs: bool) -> BackgroundSequences:
        sequence_path = PathBuilder.build(path / f"gen_model/") / f"tmp_{iteration}.tsv"

        v_genes = sorted(list(set(chain(signal.v_call for signal in sim_item.signals if signal.v_call is not None))))
        j_genes = sorted(list(set(chain(signal.j_call for signal in sim_item.signals if signal.j_call is not None))))

        if sequence_per_signal_count['no_signal'] > 0 or need_background_seqs or (
                len(v_genes) == 0 and len(j_genes) == 0) \
                or not sim_item.generative_model.can_generate_from_skewed_gene_models():
            sim_item.generative_model.generate_sequences(self._sequence_batch_size, seed=sim_item.seed,
                                                         path=sequence_path,
                                                         sequence_type=self.sequence_type,
                                                         compute_p_gen=self._use_p_gens)

            print_log(f"Generated {self._sequence_batch_size} background sequences, stored at {sequence_path}.", True)

        skew_model_for_signal = needs_seqs_with_signal(sequence_per_signal_count)

        if sim_item.generative_model.can_generate_from_skewed_gene_models() and skew_model_for_signal and (
                len(v_genes) > 0 or len(j_genes) > 0):
            sim_item.generative_model.generate_from_skewed_gene_models(v_genes=v_genes, j_genes=j_genes,
                                                                       seed=sim_item.seed, path=sequence_path,
                                                                       sequence_type=self.sequence_type,
                                                                       batch_size=self._sequence_batch_size,
                                                                       compute_p_gen=self._use_p_gens)

            print_log(
                f"Generated {self._sequence_batch_size} sequences from skewed model for given V/J genes at {sequence_path}.",
                True)

        data = get_bnp_data(sequence_path, BackgroundSequences)

        os.remove(sequence_path)
        print_log(f"Prepared sequences for processing and removed temporary file {sequence_path}.", True)

        return data

    def _make_p_gen_histogram(self, sequences: BackgroundSequences, sim_item_name: str, path: Path):

        log_p_gens = np.log10(sequences[sequences.from_default_model.astype(bool)].p_gen[
                                  np.nonzero(sequences[sequences.from_default_model.astype(bool)].p_gen)])

        hist, self.state.p_gen_bins[sim_item_name] = np.histogram(log_p_gens, density=False,
                                                                  bins=self.state.simulation.p_gen_bin_count)
        self.state.target_p_gen_histogram[sim_item_name] = hist / np.sum(sequences.from_default_model)

        zero_regions = self.state.target_p_gen_histogram[sim_item_name] == 0
        self.state.target_p_gen_histogram[sim_item_name][zero_regions] = ImplantingStrategy.MIN_RANGE_PROBABILITY
        self.state.target_p_gen_histogram[sim_item_name][np.logical_not(zero_regions)] -= \
            ImplantingStrategy.MIN_RANGE_PROBABILITY * (np.sum(zero_regions) + 1) / np.sum(np.logical_not(zero_regions))

        make_p_gen_histogram_plot(self.state.target_p_gen_histogram[sim_item_name],
                                  self.state.p_gen_bins[sim_item_name], path,
                                  str(LigoSimInstruction.MIN_RANGE_PROBABILITY))

    def _update_sequences_with_missing_p_gens(self, sequences: BackgroundSequences, sim_item: SimConfigItem,
                                              sequence_field: str = None, p_gen_field: str = "p_gen"):
        if np.any(getattr(sequences, p_gen_field) == -1):
            missing_p_gens = getattr(sequences, p_gen_field) == -1
            p_gens = getattr(sequences, p_gen_field)
            p_gens[missing_p_gens] = sim_item.generative_model.compute_p_gens(sequences[missing_p_gens], self.state.simulation.sequence_type, sequence_field)
            setattr(sequences, p_gen_field, p_gens)
        return sequences

    def _filter_using_p_gens(self, sequences: BackgroundSequences, sim_item: SimConfigItem) -> Tuple[
        BNPDataClass, dict]:
        sequences = self._update_sequences_with_missing_p_gens(sequences, sim_item)
        with np.errstate(divide='ignore'):
            p_gens = np.log10(sequences.p_gen)

        p_gen_bins = self.state.p_gen_bins[sim_item.name]
        hist = np.concatenate(
            [[LigoSimInstruction.MIN_RANGE_PROBABILITY], self.state.target_p_gen_histogram[sim_item.name],
             [LigoSimInstruction.MIN_RANGE_PROBABILITY]])

        seq_keep_prob = hist[np.digitize(p_gens, p_gen_bins)]
        keep_sequences = np.random.uniform(0, 1, len(sequences)) <= seq_keep_prob

        print_log(
            f"Removed {len(sequences) - sum(keep_sequences)} out of {len(sequences)} sequences from the batch when filtering by p_gens.",
            True)

        return sequences[keep_sequences]



import logging
import os
import shutil
from enum import Enum
from pathlib import Path

from ligo.util.PathBuilder import PathBuilder
from ligo.util.Reports import ReportOutput, ReportResult


class Util:

    @staticmethod
    def to_dict_recursive(obj, base_path: Path):
        if not hasattr(obj, "__dict__"):
            if (isinstance(obj, Path) or isinstance(obj, str)) and os.path.isfile(obj):
                obj_abs_path = os.path.abspath(obj)
                base_abs_path = os.path.abspath(str(base_path))
                res_path = os.path.relpath(obj_abs_path, base_abs_path)
                return res_path
            elif hasattr(obj, "name"):
                return obj.name
            elif isinstance(obj, list):
                return [Util.to_dict_recursive(element, base_path) for element in obj]
            else:
                return obj if obj is not None else ""
        elif isinstance(obj, Enum):
            return str(obj)
        else:
            vars_obj = vars(obj)
            result = {
                attribute_name: Util.to_dict_recursive(vars_obj[attribute_name], base_path) for attribute_name in vars_obj.keys() if not attribute_name.startswith("_")
            }
            if isinstance(obj, ReportOutput):
                path = getattr(obj, "path")

                if path is not None:
                    result['is_download_link'] = True
                    result['download_link'] = os.path.relpath(path=getattr(obj, "path"), start=base_path)
                    result['file_name'] = getattr(obj, "name")

                    if any([ext == path.suffix for ext in ['.svg', '.jpg', '.png']]):
                        result['is_embed'] = False
                    else:
                        result['is_embed'] = True
                else:
                    result['is_download_link'] = False

            return result

    @staticmethod
    def get_css_content(css_path: Path):
        with css_path.open("rb") as file:
            content = file.read()
        return content

    @staticmethod
    def make_downloadable_zip(base_path: Path, path_to_zip: Path, filename: str = "") -> str:
        if filename == "":
            filename = "_".join(Path(os.path.relpath(str(path_to_zip), str(base_path)).replace(".", "")).parts)

        PathBuilder.build(base_path / "zip")
        zip_file_path = shutil.make_archive(base_name=base_path / f"zip/{filename}", format="zip", root_dir=str(path_to_zip))
        return zip_file_path

    @staticmethod
    def get_full_specs_path(base_path) -> str:
        specs_path = list(base_path.glob("../**/full*.yaml"))

        if len(specs_path) == 1:
            path_str = os.path.relpath(specs_path[0], base_path)
        else:
            path_str = ""

        return path_str

    @staticmethod
    def get_table_string_from_csv_string(csv_string: str, separator: str = ",", has_header: bool = True) -> str:
        table_string = "<table>\n"
        for index, line in enumerate(csv_string.splitlines()):
            if index == 0 and has_header:
                table_string += "<thead>\n"
            table_string += "<tr>\n"
            for col in line.split(separator):
                table_string += f"<td>{col}</td>\n"
            table_string += "</tr>\n"
            if index == 0 and has_header:
                table_string += "</thead>\n"
        table_string += "</table>\n"
        return table_string

    @staticmethod
    def get_table_string_from_csv(csv_path: Path, separator: str = ",", has_header: bool = True) -> str:
        with csv_path.open("r") as file:
            table_string = Util.get_table_string_from_csv_string(file.read())
        return table_string

    @staticmethod
    def update_report_paths(report_result: ReportResult, path: Path) -> ReportResult:
        for attribute in vars(report_result):
            attribute_value = getattr(report_result, attribute)
            if isinstance(attribute_value, list):
                for output in attribute_value:
                    if isinstance(output, ReportOutput):
                        new_filename = "_".join([part for part in Path(os.path.relpath(path=str(output.path), start=str(path))).parts if part != ".."])
                        new_path = path / new_filename

                        if output.path != new_path:
                            shutil.copyfile(src=str(output.path), dst=new_path)
                            output.path = new_path
                    else:
                        logging.warning(f"HTML util: one of the report outputs was not returned properly from the report {report_result.name}, "
                                        f"and it will not be moved to HTML output folder.")

        return report_result



import sys
from pathlib import Path

from ligo.dsl.InstructionParser import InstructionParser
from ligo.dsl.OutputParser import OutputParser
from ligo.dsl.definition_parsers.DefinitionParser import DefinitionParser
from ligo.environment.EnvironmentSettings import EnvironmentSettings


def generate_docs(docs_path: str):
    docs_path = Path(docs_path)

    DefinitionParser.generate_docs(docs_path)
    InstructionParser.generate_docs(docs_path)
    print(f"Specification documentation is generated at {docs_path}.")


if __name__ == "__main__":
    if len(sys.argv) == 2:
        path = sys.argv[1]
    else:
        path = EnvironmentSettings.specs_docs_path
    print(path)
    generate_docs(path)



from dataclasses import dataclass
from typing import List, Union

from ligo.environment.SequenceType import SequenceType
from ligo.simulation.SimConfigItem import SimConfigItem
from ligo.simulation.simulation_strategy.SimulationStrategy import SimulationStrategy


@dataclass
class SimConfig:
    """
    Defines all parameters of the simulation.

    Arguments:

    - sim_items (dict): a list of SimConfigItems defining individual units of simulation

    - is_repertoire (bool): whether the simulation is on a repertoire (person) or sequence/receptor level

    - paired: if the simulation should output paired data, this parameter should contain a list of a list of sim_item pairs referenced by name that should be combined; if paired data is not needed, then it should be False

    - sequence_type (str): either amino_acid or nucleotide

    - simulation_strategy (str): either RejectionSampling or Implanting, see the tutorials for more information on choosing one of these

    - keep_p_gen_dist (bool): if possible, whether to keep the distribution of generation probabilities of the sequences the same as provided by the model without any signals

    - p_gen_bin_count (int): if keep_p_gen_dist is true, how many bins to use to approximate the generation probability distribution

    - remove_seqs_with_signals (bool): if true, it explicitly controls the proportions of signals in sequences and removes any accidental occurrences

    - species (str): species that the sequences come from; used to select correct genes to export full length sequences; default is 'human'

    - implanting_scaling_factor (int): determines in how many receptors to implant the signal in reach iteration; this is computed as number_of_receptors_needed_for_signal * implanting_scaling_factor; useful when using Implanting simulation strategy in combination with importance sampling, since the generation probability of some receptors with implanted signals might be very rare and those receptors might end up not being kept often with importance sampling; this parameter is only used when keep_p_gen_dist is set to True

    YAML specification:

    .. code-block:: yaml

      simulations:
        sim1:
          is_repertoire: false
          paired: false
          sequence_type: amino_acid
          simulation_strategy: RejectionSampling
          sim_items:
            sim_item1: # group of sequences with same simulation params
              generative_model:
                chain: beta
                default_model_name: humanTRB
                model_path: null
                type: OLGA
              number_of_examples: 100
              seed: 1002
              signals:
               signal1: 1

    """
    sim_items: List[SimConfigItem] = None
    identifier: str = None
    is_repertoire: bool = None
    paired: Union[bool, List[List[str]]] = None
    sequence_type: SequenceType = None
    simulation_strategy: SimulationStrategy = None
    p_gen_bin_count: int = None
    keep_p_gen_dist: bool = None
    remove_seqs_with_signals: bool = None
    species: str = None
    implanting_scaling_factor: int = None

    def __str__(self):
        return ",\n".join(str(simulation_item) for simulation_item in self.sim_items)

    def get_total_seq_count_for_signal(self, signal_id: str, model_name: str) -> int:
        sim_item_names = model_name.split("__")
        total_count = 0
        for sim_item_name in sim_item_names:
            total_count += self._get_seq_count_for_sim_item(signal_id, sim_item_name)
        return round(total_count)

    def get_total_seq_count(self, model_name: str) -> int:
        sim_item_names = model_name.split("__")
        sim_items = [sim_item for sim_item in self.sim_items if sim_item.name in sim_item_names]
        return round(
            sum([sim_item.number_of_examples * (sim_item.receptors_in_repertoire_count if self.is_repertoire else 1)
                 for sim_item in sim_items]))

    def _get_seq_count_for_sim_item(self, signal_id: str, sim_item_name: str) -> int:
        count = 0
        sim_item = [sim_item for sim_item in self.sim_items if sim_item.name == sim_item_name][0]
        signal_with_proportions = {k: v for k, v in sim_item.signal_proportions.items()
                                   if k.id == signal_id or (signal_id in k.id and "__" in k.id)}
        if signal_with_proportions:
            count = sim_item.number_of_examples * sum(signal_with_proportions.values())
            if self.is_repertoire:
                count *= sim_item.receptors_in_repertoire_count
        return count



import os
import shutil
from pathlib import Path
from typing import List

from ligo import Constants
from ligo.environment.EnvironmentSettings import EnvironmentSettings
from ligo.presentation.InstructionPresentation import InstructionPresentation
from ligo.presentation.PresentationFactory import PresentationFactory
from ligo.presentation.PresentationFormat import PresentationFormat
from ligo.presentation.TemplateParser import TemplateParser
from ligo.presentation.html.Util import Util


class HTMLBuilder:
    """
    Outputs HTML results of the analysis. This is currently the only defined format of presentation of results.

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ...
        instructions:
            ...
        output: # the output format
            format: HTML

    """

    @staticmethod
    def build(states: list, path: Path) -> Path:
        rel_path = Path(os.path.relpath(str(path)))
        presentations = HTMLBuilder._collect_all_presentations(states, rel_path)
        presentation_html_path = HTMLBuilder._make_document(presentations, rel_path)
        return presentation_html_path

    @staticmethod
    def _make_document(presentations: List[InstructionPresentation], path: Path) -> Path:
        result_path = path / "index.html"
        if len(presentations) > 1:
            html_map = {"instructions": presentations, "css_path": EnvironmentSettings.html_templates_path / "css/custom.css",
                        "full_specs": Util.get_full_specs_path(path), 'ligo_version': Constants.VERSION}
            TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "index.html",
                                 template_map=html_map, result_path=result_path)
        elif len(presentations) == 1:
            shutil.copyfile(str(presentations[0].path), str(result_path))
            HTMLBuilder._update_paths(result_path)
        else:
            result_path = None

        return result_path

    @staticmethod
    def _update_paths(result_path: Path):
        with result_path.open('r') as file:

            lines = []
            for line in file.readlines():
                if "href=" in line:
                    lines.append(line.split("href=\"")[0] + "href=\"./HTML_output/" + line.split("href=\"")[1])
                elif "src=" in line:
                    lines.append(line.split("src=\"")[0] + "src=\"./HTML_output/" + line.split("src=\"")[1])
                else:
                    lines.append(line)

        with result_path.open("w") as file:
            file.write("\n".join(lines))

    @staticmethod
    def _collect_all_presentations(states: list, rel_path: Path) -> List[InstructionPresentation]:
        presentations = []

        for state in states:
            presentation_builder = PresentationFactory.make_presentation_builder(state, PresentationFormat.HTML)
            presentation_path = presentation_builder.build(state)
            if len(states) > 1:
                presentation_path = Path(os.path.relpath(str(presentation_path), str(rel_path)))
            instruction_class = type(state).__name__[:-5]
            presentation = InstructionPresentation(presentation_path, instruction_class, state.name)
            presentations.append(presentation)

        return presentations



from pathlib import Path
from uuid import uuid4

from ligo.data_model.dataset.Dataset import Dataset
from ligo.data_model.encoded_data.EncodedData import EncodedData
from ligo.data_model.receptor.ElementGenerator import ElementGenerator


class ElementDataset(Dataset):
    """
    This is the base class for ReceptorDataset and SequenceDataset which implements all the functionality for both classes. The only difference between
    these two classes is whether paired or single chain data is stored.
    """

    def __init__(self, labels: dict = None, encoded_data: EncodedData = None, filenames: list = None, identifier: str = None,
                 file_size: int = 50000, name: str = None, element_class_name: str = None, element_ids: list = None):
        super().__init__()
        self.labels = labels
        self.encoded_data = encoded_data
        self.identifier = identifier if identifier is not None else uuid4().hex
        self.filenames = filenames if filenames is not None else []
        self.filenames = [Path(filename) for filename in self.filenames]
        self.element_generator = ElementGenerator(self.filenames, file_size, element_class_name)
        self.file_size = file_size
        self.element_ids = element_ids
        self.name = name
        self.element_class_name = element_class_name

    def get_data(self, batch_size: int = 10000):
        self.element_generator.file_list = self.filenames
        return self.element_generator.build_element_generator()

    def get_attribute(self, attribute: str):
        return list(el.get_attribute(attribute) for el in self.get_data())

    def get_batch(self, batch_size: int = 10000):
        self.element_generator.file_list = self.filenames
        return self.element_generator.build_batch_generator()

    def get_filenames(self):
        return self.filenames

    def set_filenames(self, filenames):
        self.filenames = filenames

    def get_example_count(self):
        return len(self.get_example_ids())

    def get_region_type(self):
        unique_region_types = list(set(self.get_attribute('region_type')))
        if len(unique_region_types) == 1:
            return unique_region_types[0]
        else:
            raise RuntimeError(f'Multiple region types are defined for dataset {self.name} (id={self.identifier}): '
                               f'{unique_region_types}.')

    def get_example_ids(self):
        if self.element_ids is None or (isinstance(self.element_ids, list) and len(self.element_ids) == 0):
            self.element_ids = []
            for element in self.get_data():
                self.element_ids.append(str(element.identifier))
        return self.element_ids

    def make_subset(self, example_indices, path, dataset_type: str):
        """
        Creates a new dataset object with only those examples (receptors or receptor sequences) available which were given by index in example_indices argument.

        Args:
            example_indices (list): a list of indices of examples (receptors or receptor sequences) to use in the new dataset
            path (Path): a path where to store the newly created dataset
            dataset_type (str): a type of the dataset used as a part of the name of the resulting dataset; the values are defined as constants in :py:obj:`~immuneML.data_model.dataset.Dataset.Dataset`

        Returns:

            a new dataset object (ReceptorDataset or SequenceDataset, as the original dataset) which includes only the examples specified under example_indices

        """
        new_dataset = self.__class__(labels=self.labels, file_size=self.file_size, element_class_name=self.element_generator.element_class_name)
        batch_filenames = self.element_generator.make_subset(example_indices, path, dataset_type, new_dataset.identifier)
        new_dataset.set_filenames(batch_filenames)
        new_dataset.name = f"{self.name}_split_{dataset_type.lower()}"
        return new_dataset

    def get_label_names(self):
        """Returns the list of metadata fields which can be used as labels"""
        return [label for label in list(self.labels.keys()) if label not in ['region_type', 'receptor_chains', 'organism', 'species']]

    def clone(self, keep_identifier: bool = False):
        raise NotImplementedError

    def get_data_from_index_range(self, start_index: int, end_index: int):
        return self.element_generator.get_data_from_index_range(start_index, end_index)



# quality: gold
from ligo.data_model.receptor.RegionType import RegionType
from ligo.data_model.receptor.receptor_sequence.Chain import Chain
from ligo.data_model.receptor.receptor_sequence.SequenceFrameType import SequenceFrameType


class SequenceMetadata:

    def __init__(self, v_call: str = None, j_call: str = None, chain=None, duplicate_count: int = None, frame_type: str = SequenceFrameType.IN.name,
                 region_type: str = None, cell_id: str = None, custom_params: dict = None):
        self.v_call = v_call
        self.j_call = j_call
        self.chain = Chain.get_chain(chain) if chain and isinstance(chain, str) else chain if isinstance(chain, Chain) else None
        self.duplicate_count = int(float(duplicate_count)) if isinstance(duplicate_count, str) else duplicate_count
        self.frame_type = SequenceFrameType(frame_type) if frame_type and isinstance(frame_type, str) and frame_type != 'nan' else frame_type if isinstance(frame_type, SequenceFrameType) else None
        self.region_type = RegionType(region_type) if region_type and isinstance(region_type, str) and region_type != 'nan' else region_type if isinstance(region_type, RegionType) else None
        self.cell_id = cell_id
        self.custom_params = custom_params if custom_params is not None else {}

    @property
    def v_gene(self):
        return self.v_call.split("\*")[0]

    @property
    def j_gene(self):
        return self.j_call.split("\*")[0]

    def get_attribute(self, name: str):
        """Returns the attribute value if attribute is present either directly or in custom_params, otherwise returns None"""
        if hasattr(self, name):
            return getattr(self, name)
        elif name in self.custom_params:
            return self.custom_params[name]
        else:
            return None



from pathlib import Path

from setuptools import setup, find_packages

from ligo.environment.Constants import Constants

setup(
    name="ligo",
    version=Constants.VERSION,
    description="LIgO is a tool for simulation of adaptive immune receptors and repertoires.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Milena Pavlovic",
    author_email="milenpa@student.matnat.uio.no",
    url="https://github.com/uio-bmi/ligo",
    install_requires=["pandas", "bionumpy==0.2.26", "numpy", "olga", "pyyaml", "airr<1.4", "plotly", "pystache", "scipy", 'npstructures', 'stitchr',
                      'IMGTgeneDL', 'dill'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3"
    ],
    python_requires='>=3.11',
    packages=find_packages(exclude=["test", "test.*"]),
    package_data={
        'ligo': [str(Path('IO/dataset_import/conversion/*.csv')),
                 str(Path("presentation/html/templates/*.html")),
                 str(Path("presentation/html/templates/css/*.css"))] +
                [str(Path("config/default_params/") / dir.name / "*.yaml") for dir in Path("ligo/config/default_params/").glob("*")],
    },
    entry_points={
        'console_scripts': [
            'ligo = ligo.app.LigoApp:main'
        ]
    },
)



import os
from pathlib import Path

from ligo import Constants
from ligo.environment.EnvironmentSettings import EnvironmentSettings
from ligo.presentation.TemplateParser import TemplateParser
from ligo.presentation.html.Util import Util
from ligo.simulation.LigoSimState import LigoSimState
from ligo.util.PathBuilder import PathBuilder
from ligo.util.StringHelper import StringHelper


class LIgOSimulationHTMLBuilder:
    """
    A class that will make a HTML file(s) out of SimulationState object to show what analysis took place in
    the LIgOSimulationInstruction.
    """

    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"

    @staticmethod
    def build(state: LigoSimState) -> str:
        """
        Function that builds the HTML files based on the Simulation state.
        Arguments:
            state: SimulationState object including all details of the Simulation instruction
        Returns:
             path to the main HTML file (which is located under state.result_path)
        """
        base_path = PathBuilder.build(state.result_path / "../HTML_output/")
        html_map = LIgOSimulationHTMLBuilder.make_html_map(state, base_path)
        result_file = base_path / f"Simulation_{state.name}.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "Simulation.html",
                             template_map=html_map, result_path=result_file)

        return result_file

    @staticmethod
    def make_html_map(state: LigoSimState, base_path: Path) -> dict:

        html_map = {
            "css_style": Util.get_css_content(LIgOSimulationHTMLBuilder.CSS_PATH),
            "name": state.name,
            'ligo_version': Constants.VERSION,
            "full_specs": Util.get_full_specs_path(base_path),
            "dataset_name": state.resulting_dataset.name if state.resulting_dataset.name is not None else state.resulting_dataset.identifier,
            "dataset_type": StringHelper.camel_case_to_word_string(type(state.resulting_dataset).__name__),
            "example_count": state.resulting_dataset.get_example_count(),
            "dataset_size": f"{state.resulting_dataset.get_example_count()} {type(state.resulting_dataset).__name__.replace('Dataset', 's').lower()}",
            "labels": [{"label_name": label} for label in state.resulting_dataset.get_label_names()],
            "formats": [
                {
                    "format_name": format_name,
                    "dataset_download_link": os.path.relpath(
                        path=Util.make_downloadable_zip(state.result_path, state.paths[state.resulting_dataset.name][format_name]),
                        start=base_path)
                } for format_name in state.formats
            ],
            "simulation_items": [Util.to_dict_recursive(sim_item, base_path) for sim_item in state.simulation.sim_items]
        }

        return html_map
